#if UNITY_5_3_OR_NEWER
using System.Collections;
using UnityEngine;
using UnityEngine.Networking;

namespace RoboidControl.Unity {

    /// <summary>
    /// The representation of a Thing in Unity
    /// </summary>
    public class Thing : MonoBehaviour {

        /// <summary>
        /// The core C# thing
        /// </summary>
        //[field: SerializeField]
        public RoboidControl.Thing core { get; set; }

        public SiteServer participant;

        private string modelUrl = null;

        /// <summary>
        /// Set the core C# thing
        /// </summary>
        protected void SetCoreThing(RoboidControl.Thing thing) {
            core = thing;
            core.component = this;

            SiteServer siteServer = FindAnyObjectByType<SiteServer>();
            if (siteServer == null || siteServer.site == null) {
                Debug.LogWarning("No site server found");
                return;
            }
            siteServer.site.Add(thing);
        }

        public static Thing Create(RoboidControl.Thing core) {
            // Debug.Log("Creating new Unity thing");
            GameObject gameObj = string.IsNullOrEmpty(core.name) ?
                new("Thing") :
                new(core.name);
            Thing component = gameObj.AddComponent<Thing>();
            component.Init(core);
            return component;
        }

        protected void Init(RoboidControl.Thing core) {
            this.core = core;
            this.core.component = this;
            this.participant = FindAnyObjectByType<SiteServer>();
            core.owner = this.participant.coreParticipant;
            if (core.parent != null && core.parent.component != null) {
                this.transform.SetParent(core.parent.component.transform, false);
                this.transform.localPosition = Vector3.zero;
            }

            if (core.position != null)
                this.transform.localPosition = core.position.ToVector3();
            if (core.orientation != null)
                this.transform.localRotation = core.orientation.ToQuaternion();
        }

        /// <summary>
        /// Update the Unity representation
        /// </summary>
        protected virtual void Update() {
            if (core == null)
                return;

            if (core.linearVelocity != null && core.linearVelocity.distance != 0) {
                Vector3 direction = Quaternion.AngleAxis(core.linearVelocity.direction.horizontal, Vector3.up) * Vector3.forward;
                this.transform.Translate(core.linearVelocity.distance * Time.deltaTime * direction, Space.Self);
            }

            if (core.angularVelocity != null && core.angularVelocity.distance != 0) {
                Vector3 axis = core.angularVelocity.direction.ToVector3();
                this.transform.localRotation *= Quaternion.AngleAxis(core.angularVelocity.distance * Time.deltaTime, axis);
            }
        }

        protected virtual void FixedUpdate() {
            UpdateThing();
        }

        public void UpdateThing() {
            if (core == null) {
                Debug.Log($"{this} core thing is gone, self destruct in 0 seconds...");
                Destroy(this);
                return;
            }

            if (core.updateQueue.TryDequeue(out RoboidControl.Thing.UpdateEvent e))
                HandleUpdateEvent(e);
        }

        private void HandleUpdateEvent(RoboidControl.Thing.UpdateEvent e) {
            switch (e.messageId) {
                case ThingMsg.id:
                    if (core.parent == null)
                        this.transform.SetParent(null, true);
                    else if (core.parent.component != null)
                        this.transform.SetParent(core.parent.component.transform, true);
                    break;
                case NameMsg.Id:
                    this.gameObject.name = core.name;
                    break;
                case ModelUrlMsg.Id:
                    string extension = core.modelUrl[core.modelUrl.LastIndexOf(".")..];
                    if (extension == ".jpg" || extension == ".png")
                        StartCoroutine(LoadJPG());

                    this.modelUrl = core.modelUrl;
                    break;
                case PoseMsg.Id:
                    this.HandlePose();
                    break;
                case BinaryMsg.Id:
                    this.HandleBinary();
                    break;
            }
        }

        private IEnumerator LoadJPG() {
            UnityWebRequest request = UnityWebRequestTexture.GetTexture(core.modelUrl);
            yield return request.SendWebRequest();

            if (request.result == UnityWebRequest.Result.Success) {
                Texture2D texture = ((DownloadHandlerTexture)request.downloadHandler).texture;
                float aspectRatio = (float)texture.width / (float)texture.height;

                GameObject modelQuad = GameObject.CreatePrimitive(PrimitiveType.Quad);
                Collider c = modelQuad.GetComponent<Collider>();
                c.enabled = false;
                Destroy(c);
                modelQuad.transform.SetParent(this.transform, false);
                modelQuad.transform.localEulerAngles = new(90, -90, 0);
                modelQuad.transform.localScale = new Vector3(aspectRatio, 1, 1) / 5;
                if (this.name == "Ant")
                    modelQuad.transform.localScale *= 2;
                Material quadMaterial = new(Shader.Find("Unlit/Transparent")) {
                    mainTexture = texture
                };
                modelQuad.GetComponent<Renderer>().material = quadMaterial;
            }
            else {
                Debug.LogError("Failed to load image: " + request.error);
            }
        }

        protected virtual void HandlePose() {
            if (core.linearVelocity.distance == 0)
                this.transform.localPosition = core.position.ToVector3();
            if (core.angularVelocity.distance == 0)
                this.transform.localRotation = core.orientation.ToQuaternion();

        }
        protected virtual void HandleBinary() { }

    }

}
#endif
